status_codes = {
    100: b"100 Continue",
    101: b"101 Switching Protocols",
    102: b"102 Processing",
    103: b"103 Early Hints",
    200: b"200 OK",
    201: b"201 Created",
    202: b"202 Accepted",
    203: b"203 Non-Authoritative Information",
    204: b"204 No Content",
    205: b"205 Reset Content",
    206: b"206 Partial Content",
    207: b"207 Multi-Status",
    208: b"208 Already Reported",
    226: b"226 IM Used (HTTP Delta encoding)",
    300: b"300 Multiple Choices",
    301: b"301 Moved Permanently",
    302: b"302 Found",
    303: b"303 See Other",
    304: b"304 Not Modified",
    305: b"305 Use Proxy Deprecated",
    306: b"306 unused",
    307: b"307 Temporary Redirect",
    308: b"308 Permanent Redirect",
    400: b"400 Bad Request",
    401: b"401 Unauthorized",
    402: b"402 Payment Required Experimental",
    403: b"403 Forbidden",
    404: b"404 Not Found",
    405: b"405 Method Not Allowed",
    406: b"406 Not Acceptable",
    407: b"407 Proxy Authentication Required",
    408: b"408 Request Timeout",
    409: b"409 Conflict",
    410: b"410 Gone",
    411: b"411 Length Required",
    412: b"412 Precondition Failed",
    413: b"413 Payload Too Large",
    414: b"414 URI Too Long",
    415: b"415 Unsupported Media Type",
    416: b"416 Range Not Satisfiable",
    417: b"417 Expectation Failed",
    418: b"418 I'm a teapot",
    421: b"421 Misdirected Request",
    422: b"422 Unprocessable Entity",
    423: b"423 Locked",
    424: b"424 Failed Dependency",
    425: b"425 Too Early Experimental",
    426: b"426 Upgrade Required",
    428: b"428 Precondition Required",
    429: b"429 Too Many Requests",
    431: b"431 Request Header Fields Too Large",
    451: b"451 Unavailable For Legal Reasons",
    500: b"500 Internal Server Error",
    501: b"501 Not Implemented",
    502: b"502 Bad Gateway",
    503: b"503 Service Unavailable",
    504: b"504 Gateway Timeout",
    505: b"505 HTTP Version Not Supported",
    506: b"506 Variant Also Negotiates",
    507: b"507 Insufficient Storage",
    508: b"508 Loop Detected",
    510: b"510 Not Extended",
    511: b"511 Network Authentication Required",
}
